#define PROGNAME "awriter"
#define ALSA_PCM_NEW_HW_PARAMS_API
#define BASE_ADDRESS 3758161665L /* 224.0.255.1 */
#define BASE_PORT 6000

#include <alsa/asoundlib.h>
#include <plib/socket.h>

static BOOL exit_flag = FALSE;
static char buffer[880];
static size_t size = 0;
static char place[128];

void term_signal(int signum) {
    exit_flag = TRUE;
}

void int_signal(int signum) {
    exit_flag = TRUE;
}

static void usage( void) {
    fprintf(stderr, "Usage: awriter PATH_TO_FILE IDENTIFIER [[IDENTIFIER]]\n");
    exit(1);
}

static void mix(void * dst, const void * src, size_t count) {
    REGISTER size_t i;
    REGISTER short * dst_ptr = (short *)dst, * src_ptr = (short *)src;

    for (i = 0; i < (count>>1); i ++) {
	int srcVal = *src_ptr++;
	srcVal += *dst_ptr;
	*dst_ptr++ = (short)(srcVal > 0x7fff ? 0x7fff : srcVal);
    }
}

struct hwrits {
    char * id;
    unsigned long period;
    FILE * file;
};

static void my_read( struct listen_addr * la) {
    char buf[880];
    ssize_t sz = 0;
    unsigned long tm, period;
    time_t now;
    struct hwrits * h = (struct hwrits *)la->arg;
    char filename[256];

    memset(buf, 0, sizeof(buf));
    sz = sock_read( la->fd, (void *)(buf+4), size, 10);
    if ( sz == size) {
	now = time(NULL);
	tm = (unsigned long)now;
	period = tm / 900;
	if ( h->period != period) {
	    h->period = period;
	    if ( h->file != NULL) {
		fclose(h->file);
		h->file = NULL;
	    }
	}
	if ( h->file == NULL) {
	    snprintf(filename, sizeof(filename), "%s/%s-%lu.pcmt", place, h->id, tm);
	    if ((h->file = fopen(filename, "w")) == NULL) {
		log_err("Could not open file (%s) for writing\n", filename);
		exit(1);
	    }
	}
	tm = htonl(tm);
	memcpy(buf, &tm, 4);
	if ( h->file)
	    fwrite((const void *)buf, sizeof(char), sz+4, h->file);
	mix( buffer, (void *)(buf+4), size);
    }
}

void * my_thread( void * arg) {
    while (!exit_flag) {
    	sock_loop_read(arg);
    }
    return NULL;
}

int main( int argc, char * argv[]) {
    int rc;
    snd_pcm_t *handle;
    snd_pcm_hw_params_t *params;
    unsigned int val;
    int dir = 0;
    snd_pcm_uframes_t frames;
    snd_pcm_state_t state;
    struct listen_addr * la;
    struct join_group * j;
    struct join_groups * join;
    struct listen_addrs * addrs;
    struct hwrits * h;
    POLL_T * p;
    BOOL first = TRUE;
    size_t tries, i;
    char * ptr;
    struct sockaddr_in maddr;


    plib_init();
    log_open(PROGNAME, L_INFO|L_NOTICE|L_DBG|L_ERR);
    if ( argc < 2)
	usage();
    snprintf(place, sizeof(place), "%s", argv[1]);
    i = 2;
    if ((addrs = (struct listen_addrs *)ZALLOC(sizeof(struct listen_addrs))) == NULL) {
	log_err("Could not allocate memory for listen addrs\n");
	exit(1);
    }
    addrs->tqh_last = &addrs->tqh_first;
    while ( i < argc) {
	if ((h = (struct hwrits *)ZALLOC(sizeof(struct hwrits))) == NULL) {
	    log_err("Could not allocate memory for hwrits\n");
	    exit(1);
	}
	h->id = argv[i];
	if ((join = (struct join_groups *)ZALLOC(sizeof(struct join_groups))) == NULL) {
	    log_err("Could not allocate memory for join groups\n");
	    exit(1);
	}
	join->tqh_last = &join->tqh_first;
	if ((j = (struct join_group *)ZALLOC(sizeof(struct join_group))) == NULL) {
	    log_err("Could not allocate memory for join group\n");
	    exit(1);
	}
	memset((void *)&maddr, 0, sizeof(maddr));
	maddr.sin_addr.s_addr = htonl(BASE_ADDRESS+atoi(argv[i]));
	maddr.sin_port = htons(BASE_PORT+(unsigned short)atoi(argv[i])*10);
	maddr.sin_family = AF_INET;
	addr_ston((const struct sockaddr *)&maddr, &j->addr);
	TAILQ_INSERT_TAIL( join, j, entry);
	if ((la = (struct listen_addr *)ZALLOC(sizeof(struct listen_addr))) == NULL) {
	    log_err("Could not allocate memory for listen addr\n");
	    exit(1);
	}
	la->arg = h;
	la->option = SOCK_TYPE_UDP|SOCK_OPT_RECV_MIN|SOCK_OPT_SEND_MIN;
	addr_pton("0.0.0.0", &la->addr);
	la->port = (unsigned short)BASE_PORT+(unsigned short)atoi(argv[i])*10;
	la->f = my_read;
	la->join = join;
	TAILQ_INSERT_TAIL( addrs, la, entry);
	i ++;
    }
    if ((p = sock_init_read( addrs)) == NULL)
	log_err("Could not allocate memory!");
    /* Open PCM device for recording (capture). */
    rc = snd_pcm_open(&handle, "default", SND_PCM_STREAM_PLAYBACK, 0);
    if (rc < 0)
	log_err("unable to open pcm device: %s\n", snd_strerror(rc));
    /* Allocate a hardware parameters object. */
    snd_pcm_hw_params_alloca(&params);
    /* Fill it in with default values. */
    snd_pcm_hw_params_any(handle, params);
    /* Set the desired hardware parameters. */
    /* Interleaved mode */
    snd_pcm_hw_params_set_access(handle, params, SND_PCM_ACCESS_RW_INTERLEAVED);

    /* Use a-Law format */
    snd_pcm_hw_params_set_format(handle, params, SND_PCM_FORMAT_S16_LE);
    /* One channel (mono) */
    snd_pcm_hw_params_set_channels(handle, params, 1);
    /* 8000 bits/second sampling rate (phone quality) */
    val = 8000;
    snd_pcm_hw_params_set_rate_near(handle, params, &val, &dir);
    /* Set period size to 440 frames. */
    frames = 440;
    snd_pcm_hw_params_set_period_size_near(handle, params, &frames, &dir);
    /* Write the parameters to the driver */
    rc = snd_pcm_hw_params(handle, params);
    if (rc < 0)
	log_err( "unable to set hw parameters: %s\n", snd_strerror(rc));

    /* Use a buffer large enough to hold one period */
    snd_pcm_hw_params_get_period_size(params, &frames, &dir);
    frames = 440;
    size = frames * 2;
    /* We want to loop for 5 seconds */
    snd_pcm_hw_params_get_period_time(params, &val, &dir);

    port_signal(SIGINT, int_signal);
    port_signal(SIGTERM, term_signal);
    if ((rc = snd_pcm_start(handle)) < 0) {
	if ((rc = snd_pcm_recover(handle, rc, 0)) < 0) {
	    log_debug("Start error - %s\n", snd_strerror(rc));
	    exit(1);
	}
    }
    while (!exit_flag) {
	memset( buffer, 0, sizeof(buffer));
	sock_loop_read(p);
	state = snd_pcm_state(handle);
	if ( state == SND_PCM_STATE_XRUN) {
	    if ((rc = snd_pcm_recover(handle, -EPIPE, 0)) < 0) {
		log_debug("Recover from underrun failed - %s\n", snd_strerror(rc));
		exit(1);
	    }
	    first = TRUE;
	} else if ( state == SND_PCM_STATE_SUSPENDED) {
	    if ((rc = snd_pcm_recover(handle, -ESTRPIPE, 0)) < 0) {
		log_debug("error from writen: %s\n", snd_strerror(rc));
		exit(1);
	    }
	}
	tries = 0;
	while ((rc = snd_pcm_writei(handle, buffer, frames)) < 0) {
	    if ((rc = snd_pcm_recover(handle, rc, 0)) < 0) {
		log_debug("snd_pcm_writei error - %s\n", snd_strerror(rc));
		tries ++;
	    }
	    if ( tries >= 5)
		goto out;
	}
    }
out:
    snd_pcm_drain(handle);
    snd_pcm_close(handle);
    sock_close_read( p);
    TAILQ_FOREACH(la, addrs, entry) {
	h = la->arg;
	if ( h->file)
	    fclose(h->file);
	ZFREE(h);
	TAILQ_FOREACH( j, la->join, entry) {
	    TAILQ_REMOVE(la->join, j, entry);
	    ZFREE( j);
	}
	ZFREE( la->join);
	TAILQ_REMOVE(addrs, la, entry);
	ZFREE( la);
    }
    ZFREE(addrs);
    plib_close();
    return 0;
}
